#include <ruby.h>
#include <node.h>
#include <env.h>
#include <st.h>

#include "yarvcore.h"
#include "vm.h"
#include "insnhelper.h"
#include "vm_macro.inc"
#include "insns.inc"

#define BUFSIZE 0x100

#define EVALBODY_HELPER_FUNCTION static inline

#ifndef HAVE_TYPE_ULONG
typedef unsigned long ulong;
#endif

typedef ulong GENTRY;

void stack_dump_thobj(yarv_thread_t *thobj);
void vm_analysis_operand(int insn, int n, VALUE op);
void vm_analysis_register(int reg, int isset);
void vm_analysis_insn(int insn);

VALUE th_eval_body(yarv_thread_t *th);

#if OPT_STACK_CACHING
static VALUE yarv_finish_insn_seq[1] = {BIN(finish_SC_ax_ax)};
#elif OPT_CALL_THREADED_CODE
static VALUE const yarv_finish_insn_seq[1] = {0};
#else
static VALUE yarv_finish_insn_seq[1] = {BIN(finish)};
#endif

/* imporeted from eval.c */
struct cache_entry {		/* method hash table. */
    ID mid;			/* method's id */
    ID mid0;			/* method's original id */
    VALUE klass;		/* receiver's class */
    VALUE origin;		/* where method defined  */
    NODE *method;
    int noex;
};

#define CACHE_SIZE 0x800

//RUBY_EXTERN struct cache_entry ruby_cache_table[CACHE_SIZE];
//static struct cache_entry *global_method_cache = &ruby_cache_table[0];

#define cache global_method_cache

#define CACHE_MASK 0x7ff
#define EXPR1(c,m) ((((c)>>3)^(m))&CACHE_MASK)

NODE*
rb_get_method_body(VALUE *klassp, ID *idp, int *noexp);

#include "call_cfunc.h"

/*
 * prepare stack frame
 */
static inline
VALUE th_set_env(yarv_thread_t *th, yarv_iseq_t *iseq,
                 VALUE magic, VALUE self, VALUE specval,
                 VALUE *pc, VALUE *sp, VALUE *lfp,
                 int local_size, int argc, const VALUE *argv){
  VALUE *dfp;
  int i;
 
  for(i=0; i<argc; i++){
    *sp = *argv;
    sp++; argv++;
  }

  for(; i<local_size; i++){
    *sp = Qnil;
    sp++;
  }
  
  *sp = GC_GUARDED_PTR(specval);
  dfp = sp;
  if(lfp == 0){
    lfp = sp;
  }
  PUSH_CONTROL_STACK_FRAME(th,
                           pc,            // pc
                           sp+1,          // sp
                           sp+1,          // bp
                           iseq,          // iseq
                           magic,         // magic
                           self,          // self
                           lfp,           // lfp
                           dfp            // dfp
                           );

  return Qtrue;
}

EXTERN VALUE ruby_top_self;

VALUE th_set_finish_env(yarv_thread_t *th){
  th_set_env(th, 0,
             FRAME_MAGIC_FINISH, Qnil, 0,
             0, th->cfp->sp, 0,
             0, 0, 0);
  th->cfp->pc = &yarv_finish_insn_seq[0];
  return Qtrue;
}


VALUE thread_set_top_stack(VALUE self, VALUE iseqval){
  yarv_iseq_t *iseq;
  yarv_thread_t *th;

  GetThreadVal(self, th);
  GetISeqVal(iseqval, iseq);
  
  iseq->klass_nest_stack = rb_ary_push(rb_ary_new(), rb_cObject);

  /* for return */
  th_set_finish_env(th);
  
  return th_set_env(th, iseq, 
                    FRAME_MAGIC_TOP, ruby_top_self, 0,
                    iseq->iseq_encoded, th->cfp->sp, 0,
                    iseq->local_size, 0, 0);
}

static
VALUE th_make_env_object(yarv_thread_t *th, yarv_control_frame_t *cfp,
                         VALUE *envptr, VALUE *endptr){
  VALUE envval, penvval = 0;
  yarv_env_t *env;
  VALUE *nenvptr;
  int i, local_size;
  
  if(ENV_IN_HEAP_P(envptr)){
    return ENV_VAL(envptr);
  }

  if(envptr != endptr){
    VALUE *penvptr = GC_GUARDED_PTR_REF(*envptr);
    yarv_control_frame_t *pcfp = cfp;
    while(pcfp->dfp != penvptr){
      pcfp++;
    }
    penvval = th_make_env_object(th, pcfp, penvptr, endptr);
    cfp->lfp = pcfp->lfp;
    *envptr = GC_GUARDED_PTR(pcfp->dfp);
  }
  
  /* allocate env */
  envval = rb_obj_alloc(cYarvEnv);
  GetEnvVal(envval, env);

  local_size = cfp->iseq->local_size;
  env->env_size   = local_size + 1 + 4;
  env->local_size = local_size;
  env->env = ALLOC_N(VALUE, env->env_size);
  env->prev_envval = penvval;

  for(i=0; i<=local_size; i++){
    env->env[i] = envptr[-local_size + i];
    envptr[-local_size + i] = 0;
  }
  *envptr = envval; /* GC mark */
  
  nenvptr = &env->env[i-1];
  nenvptr[1] = Qfalse; /* frame is not orphan */
  nenvptr[2] = Qundef; /* frame is in heap    */
  nenvptr[3] = envval; /* frame self */
  nenvptr[4] = penvval;/* frame prev env object */
  
  /* reset lfp/dfp in cfp */
  cfp->dfp = nenvptr;
  if(envptr == endptr){
    cfp->lfp = nenvptr;
  }
  return envval;
}

VALUE th_make_proc(yarv_thread_t *th, yarv_control_frame_t *cfp, yarv_block_t *block);

static
VALUE th_make_proc_from_block(yarv_thread_t *th, yarv_control_frame_t *cfp,
                              yarv_block_t *block){
  VALUE procval;
  yarv_control_frame_t *bcfp;
  VALUE *bdfp; /* to gc mark */
  
  if(block->proc){
    return block->proc;
  }
  
  bcfp = GET_CFP_FROM_BLOCK_PTR(block);
  bdfp = bcfp->dfp;
  procval = th_make_proc(th, bcfp, block);
  bdfp[-1] = procval;
  return procval;
}

VALUE th_make_proc(yarv_thread_t *th, yarv_control_frame_t *cfp, yarv_block_t *block){
  VALUE procval, envval;
  yarv_proc_t *proc;

  procval = rb_obj_alloc(cYarvProc);
  GetProcVal(procval, proc);

  if(GC_GUARDED_PTR_REF(*cfp->lfp) != 0){
    yarv_proc_t *p;
    proc->blockprocval = th_make_proc_from_block(th, cfp, (yarv_block_t *)GC_GUARDED_PTR_REF(*cfp->lfp));
    GetProcVal(proc->blockprocval, p);
    *cfp->lfp = GC_GUARDED_PTR(&p->block);
  }

  envval = th_make_env_object(th, cfp, cfp->dfp, cfp->lfp);
  
  proc->block.self = block->self;
  proc->block.lfp  = block->lfp;
  proc->block.dfp  = block->dfp;
  proc->block.iseq = block->iseq;
  proc->block.proc = procval;
  proc->envval      = envval;

  return procval;
}


VALUE th_call0(yarv_thread_t *th, VALUE klass, VALUE recv,
               VALUE id, ID oid, int argc, const VALUE *argv,
               NODE *body, int nosuper){
  VALUE val;
  yarv_block_t *blockptr = 0;

  //printf("id: %s\n", rb_id2name(id));
  //SDR2(th->cfp);
  
  if(th->ifuncnode){
    blockptr = GET_BLOCK_PTR_IN_CFP(th->cfp);
    blockptr->iseq = (void *)th->ifuncnode;
    blockptr->proc = 0;
    th->ifuncnode = 0;
  }
  
  switch(nd_type(body)){
  case YARV_METHOD_NODE:{
    yarv_control_frame_t *reg_cfp;
    int i;
    const int flag = 0;
    
    th_set_finish_env(th);
    
    reg_cfp = th->cfp;
    
    for(i=0; i<argc; i++){
      *reg_cfp->sp++ = argv[i];
    }

    macro_eval_invoke_func(body->nd_body, recv, klass, blockptr, argc);
    val = th_eval_body(th);
    break;
  }
  case NODE_CFUNC:{
    yarv_control_frame_t *reg_cfp = th->cfp;
    struct cmethod_info cmi = {0, id, klass};

    th_set_env(th, (yarv_iseq_t *)&cmi,
               FRAME_MAGIC_CFUNC, recv, (VALUE)blockptr,
               0, reg_cfp->sp, 0,
               0, 0, 0);
    val = call_cfunc(body->nd_cfnc, recv, body->nd_argc, argc, argv);
    th->cfp = reg_cfp; /* pop control stack frame */
    break;
  }
  case NODE_ATTRSET:{
    if (argc != 1){
      rb_raise(rb_eArgError, "wrong number of arguments (%d for 1)", argc);
    }
    val = rb_ivar_set(recv, body->nd_vid, argv[0]);
    break;
  }
  case NODE_IVAR:{
    if (argc != 0) {
      rb_raise(rb_eArgError, "wrong number of arguments (%d for 0)", argc);
    }
    val = rb_attr_get(recv, body->nd_vid);
    break;
  }
  default:
    rb_bug("unsupported: thread_call0");
  }
  return val;
}

VALUE thread_call0(VALUE self, VALUE klass, VALUE recv, VALUE id, ID oid,
                   int argc, const VALUE *argv, NODE *body, int nosuper){
  yarv_thread_t *th;
  GetThreadVal(self, th);
  return th_call0(th, klass, recv, id, oid, argc, argv, body, nosuper);
}

VALUE th_call_super(yarv_thread_t *th, int argc, const VALUE *argv){
  VALUE recv  = th->cfp->self;
  VALUE klass;
  ID id;
  NODE *body;
  int nosuper = 0;
  {
    struct cmethod_info *cmi = (void *)th->cfp->iseq;
    if(cmi->sig == 0){
      klass = RCLASS(cmi->klass)->super;
      id = cmi->id;
    }
    else{
      rb_bug("thread_call_super: should not be reached");
    }
  }
  
  body = rb_method_node(klass, id);
  return th_call0(th, klass, recv, id, id, argc, argv, body, nosuper);
}


VALUE thread_call_super(VALUE self, int argc, const VALUE *argv){
  yarv_thread_t *th;
  GetThreadVal(self, th);
  return th_call_super(th, argc, argv);
}


static inline
VALUE thread_invoke_yield_cfunc(yarv_thread_t *th,
                                yarv_block_t *block, int argc, VALUE *argv){

  NODE *ifunc = (NODE*)block->iseq;
  VALUE val;
  VALUE arg;

  if(argc == 1){
    arg = *argv;
  }
  else if(argc > 1){
    arg = rb_ary_new4(argc, argv);
  }

  th_set_env(th, 0,
             FRAME_MAGIC_IFUNC, block->self, (VALUE)block->dfp,
             0, th->cfp->sp, block->lfp,
             0, 0, 0);

  val = (*ifunc->nd_cfnc)(arg, ifunc->nd_tval, Qnil);
  
  th->cfp++;
  return val;
}


VALUE th_invoke_yield(yarv_thread_t *th, int argc, VALUE *argv){
  yarv_control_frame_t *cfp = th->cfp;
  yarv_block_t *block = GC_GUARDED_PTR_REF(cfp->lfp[0]);
  VALUE val;

  if(block == 0){
    rb_bug("block is not given");
  }
  else{
    if(BUILTIN_TYPE(block->iseq) != T_NODE){
      th_set_finish_env(th);
      th_set_env(th, block->iseq, FRAME_MAGIC_BLOCK,
                 block->self, GC_GUARDED_PTR(block->dfp),
                 block->iseq->iseq_encoded, th->cfp->sp, block->lfp,
                 block->iseq->local_size, argc, argv);
      //SDR2(th->cfp);
      val = th_eval_body(th);
    }
    else{
      val = thread_invoke_yield_cfunc(th, block, argc, argv);
    }
  }
  return val;
}


VALUE thread_yield_light_prepare(VALUE self, int argc, VALUE *argv,
                                 struct yarv_yield_data *data){
  yarv_thread_t *th;
  yarv_control_frame_t *cfp;
  yarv_block_t *block;

  GetThreadVal(self, th);
  cfp = th->cfp;
  
  data->th    = th;
  data->block = block = GC_GUARDED_PTR_REF(cfp->lfp[0]);

  if(block == 0){
    rb_bug("block is not given");
  }
  else{
    if(BUILTIN_TYPE(block->iseq) != T_NODE){
      int local_size = block->iseq->local_size;
      th_set_finish_env(th);

      if(block->iseq->argc < argc){
        argc = block->iseq->argc;
      }

      th_set_env(th, block->iseq, FRAME_MAGIC_BLOCK,
                 block->self, GC_GUARDED_PTR(block->dfp),
                 block->iseq->iseq_encoded, th->cfp->sp, block->lfp,
                 local_size, argc, argv);

      data->local_size = block->iseq->local_size;
      data->pc = block->iseq->iseq_encoded;
      data->sp = th->cfp->sp;
      data->argc = argc;
      
      th->cfp+=2; // pop, pop
    }
    else{
      rb_bug("thread_yield_light_prepare: unsupported ifunc");
      //val = thread_invoke_yield_cfunc(th, block, argc, argv);
    }
  }
  return Qtrue;
}


VALUE thread_yield_light_invoke(int argc, VALUE *argv,
                                struct yarv_yield_data *data){
  VALUE val;
  yarv_thread_t *th = data->th;
  int i;
  
  if(BUILTIN_TYPE(data->block->iseq) != T_NODE){
    th->cfp -= 2;
    th->cfp->pc = data->pc;
    // th->cfp->sp = data->sp;
    
    for(i=0; i<data->argc; i++){
      th->cfp->dfp[i - data->local_size] = argv[i];
    }

    val = th_eval_body(th);
    (th->cfp-1)->pc--;
  }
  else{
    val = Qundef;
  }
  return val;
}


VALUE th_invoke_proc(yarv_thread_t *th, yarv_proc_t *proc, int argc, VALUE *argv){
  th_set_finish_env(th);
  th_set_env(th, proc->block.iseq,
             FRAME_MAGIC_PROC, proc->block.self, (VALUE)proc->block.dfp,
             proc->block.iseq->iseq_encoded, th->cfp->sp, proc->block.lfp,
             proc->block.iseq->local_size, argc, argv);

  //proc_dump_raw(proc);
  //SDR2(th->cfp);
  
  return th_eval_body(th);
}


VALUE thread_invoke_proc_call(VALUE self, VALUE procval, int argc, VALUE *argv){
  yarv_thread_t *th;
  yarv_proc_t *proc;
  GetThreadVal(self, th);
  GetProcVal(procval, proc);
  
  return th_invoke_proc(th, proc, argc, argv);
}


VALUE *th_svar(yarv_thread_t *th, int cnt){
  yarv_control_frame_t *cfp = th->cfp;
  NODE *node;

  while(cfp->pc == 0){
    cfp++;
  }
  
  node = (NODE*)cfp->lfp[-1];

  if((VALUE)node == Qnil){
    cfp->lfp[-1] = (VALUE)(node = NEW_IF(Qnil, Qnil, 0));
  }
  if(cnt == 0){
    return &node->u1.value;
  }
  else if(cnt == 1){
    return &node->u2.value;
  }
  else{
    VALUE ary;
    if((ary = node->u3.value) == 0){
      ary = node->u3.value = rb_ary_new();
    }
    if(RARRAY(ary)->len <= cnt){
      rb_ary_store(ary, cnt, Qnil);
    }
    return &RARRAY(ary)->ptr[cnt];
  }
}


VALUE *thread_svar(VALUE self, int cnt){
  yarv_thread_t *th;
  GetThreadVal(self, th);
  return th_svar(th, cnt);
}

static
VALUE th_backtrace_each(yarv_thread_t *th, yarv_control_frame_t *cfp,
                        char *file, int line_no, VALUE ary){
  VALUE str = 0;
  char buf[BUFSIZE];

  if(cfp->iseq != 0){
    if(cfp->pc != 0){
      int i;
      int pos = cfp->pc - cfp->iseq->iseq_encoded;

      for(i=0; i<cfp->iseq->insn_info_size; i++){
        if(cfp->iseq->insn_info_tbl[i].position == pos){
          line_no = cfp->iseq->insn_info_tbl[i-1].line_no;
          // sendpos = iseq->insn_info_tbl[i-1].position;
          goto found;
        }
      }
      rb_bug("thread_backtrace: unkown instruction (%d)", pos);
    found:
      snprintf(buf, BUFSIZ, "%s:%d:in `%s'",
               file = RSTRING(cfp->iseq->file_name)->ptr,
               line_no, RSTRING(cfp->iseq->name)->ptr);
    }
    else{
      struct cmethod_info *cmi = (void *)cfp->iseq;
      snprintf(buf, BUFSIZE, "%s:%d:in `%s'",
               file, line_no, rb_id2name(cmi->id));
    }
    str = rb_str_new2(buf);
  }
  else if(0 && cfp->magic == FRAME_MAGIC_IFUNC){
    snprintf(buf, BUFSIZE, "::in `<ifunc>'", // ""%s:%d:in `<ifunc>'",
             file, line_no);
    str = rb_str_new2(buf);
  }
  
  if(th->cfp != cfp){
    th_backtrace_each(th, cfp-1, file, line_no, ary);
  }
  if(str){
    rb_ary_push(ary, str);
  }
  return str;
}

VALUE th_backtrace(yarv_thread_t *th, int lev){
  VALUE ary;
  yarv_control_frame_t *cfp = th->cfp;
  yarv_control_frame_t *top_of_cfp = (void *)(th->stack + th->stack_size);
  top_of_cfp--;

  // printf("%d\n", lev);

  if(lev < 0){
    /* TODO ?? */
    ary = rb_ary_new();
  }
  else{
    while(lev-->=0){
      cfp++;
      if(cfp >= top_of_cfp){
        return Qnil;
      }
    }
    ary = rb_ary_new();
  }
  
  th_backtrace_each(th, top_of_cfp, "", 0, ary);
  return ary;
}

VALUE thread_backtrace(VALUE self, int level){
  yarv_thread_t *th;
  GetThreadVal(self, th);
  return th_backtrace(th, level);
}



/*
 *
 * vm main loop helper functions
 *
 */


EVALBODY_HELPER_FUNCTION
VALUE eval_get_ev_const(yarv_thread_t *th,
                        yarv_iseq_t *iseq, VALUE klass, ID id){
  VALUE val;
  
  if(klass == Qnil){
    /* in current lexical scope */
    VALUE cref; /* Array */
    int i, last_pos;

    cref = iseq->klass_nest_stack;
    if(cref == 0){
      /* in module or class scope */
      cref = th->klass_nest_stack;
    }
    last_pos = RARRAY(cref)->len-1;
    for(i = last_pos; i>0; i--){
      klass = RARRAY(cref)->ptr[i];
    search_continue:
      if(RCLASS(klass)->iv_tbl &&
         st_lookup(RCLASS(klass)->iv_tbl, id, &val)){
        if(val == Qundef){
          rb_autoload_load(klass, id);
          goto search_continue;
        }
        else{
          break;
        }
      }
    }
    if(i==0){
      klass = RARRAY(cref)->ptr[last_pos];
      val = rb_const_get(klass, id);
    }
  }
  else{
    val = rb_const_get(klass, id);
  }
  return val;
}

EVALBODY_HELPER_FUNCTION
int eval_get_ev_defined(yarv_thread_t *th,
                        yarv_iseq_t *iseq, VALUE klass, ID id){
  VALUE val;
  if(klass == Qnil){
    VALUE cref;
    int i, last_pos;

    if((cref = iseq->klass_nest_stack) == 0){
      cref = th->klass_nest_stack;
    }
    last_pos = RARRAY(cref)->len-1;
    for(i = last_pos; i>0; i--){
      klass = RARRAY(cref)->ptr[i];
    search_continue:
      if(RCLASS(klass)->iv_tbl &&
         st_lookup(RCLASS(klass)->iv_tbl, id, &val)){
        if(val == Qundef){
          rb_autoload_load(klass, id);
          goto search_continue;
        }
        else{
          return 1;
        }
      }
    }
    klass = RARRAY(cref)->ptr[last_pos];
    return rb_const_defined(klass, id);
  }
  else{
    return rb_const_defined(klass, id);
  }
}


EVALBODY_HELPER_FUNCTION
NODE* eval_method_search(VALUE id, VALUE klass, IC ic){
  NODE *mn;

//#undef  INLINE_METHOD_CACHE
//#define FAKE_INLINE_METHOD_CACHE
//#define FAKE_INLINE_METHOD_CACHE2
//#define INLINE_METHOD_CACHE
//#define ONLY_GLOBAL_METHOD_CACHE

#if 0
1  error
#elif defined(FAKE_INLINE_METHOD_CACHE)
  {
    /* don't support polymorphic method call */
    if(LIKELY(ic->ic_method != 0)){
      mn = ic->ic_method;
    }
    else{
      ic->ic_method = mn = rb_method_node(klass, id);
    }
  }
#elif defined(FAKE_INLINE_METHOD_CACHE2)
  {
    if(LIKELY(klass == ic->ic_klass)){
      mn = ic->ic_method;
    }
    else{
      mn = rb_method_node(klass, id);
      ic->ic_klass  = klass;
      ic->ic_method = mn;
    }
  }
#elif defined(OPT_INLINE_METHOD_CACHE)
  {
    if(LIKELY(klass == ic->ic_klass) &&
       LIKELY(GET_VM_STATE_VERSION() == ic->ic_vmstat)){
      mn = ic->ic_method;
    }
    else{
      mn = rb_method_node(klass, id);
      ic->ic_klass  = klass;
      ic->ic_method = mn;
      ic->ic_vmstat = GET_VM_STATE_VERSION();
    }
  }

#elif defined(ONLY_GLOBAL_METHOD_CACHE)
  {
    struct cache_entry *ent = cache + EXPR1(klass, id);
    int noex;
    
    if (LIKELY(ent->mid == id) && LIKELY(ent->klass == klass)){
      if(LIKELY(ent->method != 0)){
        // klass = ent->origin;
        // id    = ent->mid0;
        // noex  = ent->noex;
        mn    = ent->method;
      }
      else{
        mn = 0;
      }
    }
    else{
      mn = rb_method_node(klass, id);
    }
  }
#else
  mn = rb_method_node(klass, id);
#endif
  return mn;
}


/*********************************************************/
/*********************************************************/
/* import from ruby1.9/eval.c                            */

#include <stdio.h>
#if defined(HAVE_GETCONTEXT) && defined(HAVE_SETCONTEXT)
#include <ucontext.h>
#define USE_CONTEXT
#else
#include <setjmp.h>
#endif

#ifdef USE_CONTEXT
typedef struct {
    ucontext_t context;
    volatile int status;
} rb_jmpbuf_t[1];

#undef longjmp
#undef setjmp
NORETURN(static void rb_jump_context(rb_jmpbuf_t, int));
static inline void
rb_jump_context(env, val)
    rb_jmpbuf_t env;
    int val;
{
    env->status = val;
    setcontext(&env->context);
    abort();			/* ensure noreturn */
}
#define longjmp(env, val) rb_jump_context(env, val)
#define setjmp(j) ((j)->status = 0, getcontext(&(j)->context), (j)->status)
#else
typedef jmp_buf rb_jmpbuf_t;
#ifndef setjmp
#ifdef HAVE__SETJMP
#define setjmp(env) _setjmp(env)
#define longjmp(env,val) _longjmp(env,val)
#endif
#endif
#endif

struct tag {
    rb_jmpbuf_t buf;
    struct FRAME *frame;
    struct iter *iter;
    VALUE tag;
    VALUE retval;
    struct SCOPE *scope;
    VALUE dst;
    struct tag *prev;
    int blkid;
};

#if YARVEXT
RUBY_EXTERN struct tag *ruby_prot_tag;
#else
RUBY_EXTERN struct tag *ruby_prot_tag;
#endif

#define PUSH_TAG(ptag) do {            \
  _tag.prev = ruby_prot_tag;           \
  _tag.iter = ruby_prot_tag->iter;     \
  _tag.frame = ruby_frame;             \
  _tag.tag  = ptag;                    \
  ruby_prot_tag = &_tag;               \
} while(0)

#define POP_TAG() do {                 \
  ruby_prot_tag = ruby_prot_tag->prev; \
} while(0)

#define EXEC_TAG()                     \
  (FLUSH_REGISTER_WINDOWS, setjmp(ruby_prot_tag->buf))

#define JUMP_TAG(st) do {              \
  longjmp(ruby_prot_tag->buf,(st));    \
} while (0)

#define TAG_RETURN	0x1
#define TAG_BREAK	0x2
#define TAG_NEXT	0x3
#define TAG_RETRY	0x4
#define TAG_REDO	0x5
#define TAG_RAISE	0x6
#define TAG_THROW	0x7
#define TAG_FATAL	0x8
#define TAG_MASK	0xf

/*********************************************************/
/*********************************************************/

NORETURN(static void localjump_error(const char*, VALUE, int));
static void
localjump_error(mesg, value, reason)
    const char *mesg;
    VALUE value;
    int reason;
{
  VALUE exc = rb_exc_new2(rb_const_get(rb_cObject, rb_intern("LocalJumpError")), mesg);
  ID id;

  rb_iv_set(exc, "@exit_value", value);
  switch (reason) {
  case TAG_BREAK:
    id = rb_intern("break"); break;
  case TAG_REDO:
    id = rb_intern("redo"); break;
  case TAG_RETRY:
    id = rb_intern("retry"); break;
  case TAG_NEXT:
    id = rb_intern("next"); break;
  case TAG_RETURN:
    id = rb_intern("return"); break;
  default:
    id = rb_intern("noreason"); break;
  }
  rb_iv_set(exc, "@reason", ID2SYM(id));
  rb_exc_raise(exc);
}


VALUE yarv_unredefined_flag = 0x01;

#include "vm_evalbody.h"

/*                  finish
  VMe (h1)          finish
    VM              finish F1 F2
      func          finish F1 F2 C1
        rb_funcall  finish F1 F2 C1
          VMe       finish F1 F2 C1
            VM      finish F1 F2 C1 F3

  F1 - F3 : pushed by VM
  C1      : pushed by send insn (CFUNC)

  struct CONTROL_FRAME{
    VALUE *pc;                  // cfp[0]
    VALUE *sp;                  // cfp[1]
    VALUE *bp;                  // cfp[2]
    yarv_iseq_t *iseq;          // cfp[3]
    VALUE magic;                // cfp[4]
    VALUE self;                 // cfp[5]
    VALUE *lfp;                 // cfp[6]
    VALUE *dfp;                 // cfp[7]
    yarv_iseq_t * block_iseq;   // cfp[8]
    VALUE proc;                 // cfp[9] always 0
  };

  struct BLOCK{
    VALUE self;
    VALUE *lfp;
    VALUE *dfp;
    yarv_iseq_t *block_iseq;
  };

  struct PROC{
    VALUE  proc_sig = 0;
    struct BLOCK;
  };

  struct METHOD_CONTROL_FRAME{
    struct CONTROL_FRAME;
  };

  struct METHOD_FRAME{
    VALUE arg0;
    ...
    VALUE argM;
    VALUE param0;
    ...
    VALUE paramN;
    VALUE special;                         // lfp [1]
    struct block_object *block_ptr | 0x01; // lfp [0]
  };

  struct BLOCK_CONTROL_FRAME{
    struct STACK_FRAME;
  };

  struct BLOCK_FRAME{
    VALUE arg0;
    ...
    VALUE argM;
    VALUE param0;
    ...
    VALUE paramN;
    VALUE *(prev_ptr | 0x01); // DFP[0]
  };

  struct CLASS_CONTROL_FRAME{
    struct STACK_FRAME;
  };

  struct CLASS_FRAME{
    VALUE param0;
    ...
    VALUE paramN;
    VALUE dummy;
  };

  struct C_METHOD_CONTROL_FRAME{
    VALUE *pc;                       // 0
    VALUE *sp;                       // stack pointer
    VALUE *bp;                       // base pointer (used in exception)
    yarv_iseq_t *iseq;               // cmi
    VALUE magic;                     // C_METHOD_FRAME
    VALUE self;                      // ?
    VALUE *lfp;                      // lfp
    VALUE *dfp;                      // == lfp
    yarv_iseq_t * block_iseq;        //
    VALUE proc;                      // always 0
  };

  struct C_BLOCK_CONTROL_FRAME{
    VALUE *pc;                       // finish wƂDreturn 邾D
    VALUE *sp;                       // sp
    yarv_iseq_t *iseq;        // ?
    VALUE magic;                     // C_METHOD_FRAME
    VALUE self;                      // H@ςv邩ȁH
    VALUE *lfp;                      // lfp
    VALUE *dfp;                      // lfp
    yarv_iseq_t * block_iseq; // 0
  };
  
  struct C_METHDO_FRAME{
    VALUE block_ptr;
    VALUE special;
  };
 */


VALUE th_eval_body(yarv_thread_t *th){
  struct tag _tag;
  int state;
  VALUE result, err;
  VALUE initial = 0;
  
  PUSH_TAG(0);
  if((state = EXEC_TAG()) == 0){
  vm_loop_start:
    result = th_eval(th, initial);
    if((state = th->state) != 0){
      err = result;
      th->state = 0;
      goto exception_handler;
    }
  }
  else{
    int i;
    struct catch_table_entry *entry;
    ulong  epc, cont_pc, cont_sp;
    VALUE catch_iseqval;
    yarv_control_frame_t *cfp;
    VALUE  *escape_dfp;
    VALUE type;
    
    err = ruby_errinfo;
    if(state == TAG_RAISE){
      rb_ivar_set(err, idThrowState, state);
    }
    
  exception_handler:
    cont_pc = cont_sp = catch_iseqval = 0;
    while(th->cfp->pc == 0 || th->cfp->iseq == 0){
      th->cfp++;
    }
    cfp = th->cfp;
    epc = cfp->pc - cfp->iseq->iseq_encoded;

    if(state == TAG_BREAK || state == TAG_RETURN){
      escape_dfp = GET_THROWOBJ_CATCH_POINT(err);

      // printf("%p, %p\n", escape_dfp, cfp->dfp);

      if(cfp->dfp == escape_dfp){
        if(state == TAG_RETURN){
          SET_THROWOBJ_CATCH_POINT(err, (VALUE)(cfp+1)->dfp);
          SET_THROWOBJ_STATE      (err, state = TAG_BREAK);
          /* through */
        }
        else{
          /* TAG_BREAK */
#if OPT_STACK_CACHING
          initial = (GET_THROWOBJ_VAL(err));
#else
          *th->cfp->sp++ = (GET_THROWOBJ_VAL(err));
#endif
          ruby_errinfo = Qnil;
          goto vm_loop_start;
        }
      }
    }

    if(state == TAG_RAISE){
      for(i=0; i<cfp->iseq->catch_table_size; i++){
        entry = &cfp->iseq->catch_table[i];
        if(entry->start <  epc &&
           entry->end   >= epc){

          if(entry->type == CATCH_TYPE_RESCUE ||
             entry->type == CATCH_TYPE_ENSURE){
            catch_iseqval = entry->iseq;
            cont_pc    = entry->cont;
            cont_sp    = entry->sp;
            break;
          }
        }
      }
    }
    else if(state == TAG_RETRY){
      for(i=0; i<cfp->iseq->catch_table_size; i++){
        entry = &cfp->iseq->catch_table[i];
        if(entry->start <  epc &&
           entry->end   >= epc){
          
          if(entry->type == CATCH_TYPE_ENSURE){
            catch_iseqval = entry->iseq;
            cont_pc    = entry->cont;
            cont_sp    = entry->sp;
            break;
          }
          else if(entry->type == CATCH_TYPE_RETRY){
            VALUE  *escape_dfp;
            escape_dfp = GET_THROWOBJ_CATCH_POINT(err);

            if(cfp->dfp == escape_dfp){
              cfp->pc = cfp->iseq->iseq_encoded + entry->cont;
              goto vm_loop_start;
            }
          }
        }
      }
    }
    else if(state == TAG_BREAK &&
            ((VALUE)escape_dfp & ~0x03) == 0){
      type = CATCH_TYPE_BREAK;

    search_restart_point:
      for(i=0; i<cfp->iseq->catch_table_size; i++){
        entry = &cfp->iseq->catch_table[i];
        if(entry->start <  epc &&
           entry->end   >= epc){
          
          if(entry->type == CATCH_TYPE_ENSURE){
            catch_iseqval = entry->iseq;
            cont_pc    = entry->cont;
            cont_sp    = entry->sp;
            break;
          }
          else if(entry->type == type){
            cfp->pc = cfp->iseq->iseq_encoded + entry->cont;
            cfp->sp = cfp->bp + entry->sp;
            
            if(state != TAG_REDO && escape_dfp){
#if OPT_STACK_CACHING
              initial = (GET_THROWOBJ_VAL(err));
#else
              *th->cfp->sp++ = (GET_THROWOBJ_VAL(err));
#endif
            }
            ruby_errinfo = Qnil;
            goto vm_loop_start;
          }
        }
      }
    }
    else if(state == TAG_REDO){
      type = CATCH_TYPE_REDO;
      goto search_restart_point;
    }
    else if(state == TAG_NEXT){
      type = CATCH_TYPE_NEXT;
      goto search_restart_point;
    }
    else{
      for(i=0; i<cfp->iseq->catch_table_size; i++){
        entry = &cfp->iseq->catch_table[i];
        if(entry->start <  epc &&
           entry->end   >= epc){
          
          if(entry->type == CATCH_TYPE_ENSURE){
            catch_iseqval = entry->iseq;
            cont_pc    = entry->cont;
            cont_sp    = entry->sp;
            break;
          }
        }
      }
    }

    if(catch_iseqval != 0){
      /* found catch table */
      yarv_iseq_t* catch_iseq;
      VALUE *sp;
      
      /* enter catch scope */
      GetISeqVal(catch_iseqval, catch_iseq);
      cfp->sp = cfp->bp + cont_sp;

      /* PUSH(throwed) */
      *cfp->sp++ = err;
      cfp->pc = cfp->iseq->iseq_encoded + cont_pc;

      /* push block frame */
      th_set_env(th, catch_iseq,
                 FRAME_MAGIC_BLOCK, cfp->self, GC_GUARDED_PTR(cfp->dfp),
                 catch_iseq->iseq_encoded, cfp->sp, cfp->lfp,
                 catch_iseq->local_size - 1, 0, 0);

      state = 0;
      ruby_errinfo = Qnil;
      goto vm_loop_start;
    }
    else{
      /* pop cref */
      if(cfp->iseq->type == ISEQ_TYPE_CLASS){
        rb_ary_pop(th->klass_nest_stack);
      }
      
      th->cfp++;
      if(th->cfp->pc != &yarv_finish_insn_seq[0]){
        goto exception_handler;
      }
      else{
        th->cfp++;
        ruby_errinfo = err;
        POP_TAG();
        JUMP_TAG(state);
      }
    }
  }
  POP_TAG();
  return result;
}


VALUE thread_eval(VALUE self){
  yarv_thread_t *thobj;
  yarv_iseq_t *iseq;
  
  GetThreadVal(self , thobj);
  return th_eval_body(thobj);
}

